#include<stdio.h>
#include<stdlib.h>
#include<sys/types.h>
#include<unistd.h>
#include<pthread.h>
#include"errno.h"
#include"mpi.h"

// tag -14 is used for allgather.
int
MPI_Allgather (void *sendbuf, int sendcount, MPI_Datatype sendtype,
               void *recvbuf, int recvcount, MPI_Datatype recvtype,
               MPI_Comm comm)
{
  int nump, rank, s, i;
  int bm, nb, cb;
  int partner, cont;
  int scount, rcount, sendind, recvind;
  MPI_Status status;

  MPI_Comm_size (comm, &nump);
  MPI_Comm_rank (comm, &rank);

  if (rank == -1) {
    printf ("Wrong communicator in calling MPI_Allgather\n");
    fflush (stdout);
    return 1;
  }


  s = _USF_mpi_sizeof (recvtype);
  nb = _USF_num_bits (nump - 1);
  bm = _USF_make_bm (nb);


  memcpy ((recvbuf + (rank * recvcount * s)), sendbuf, recvcount * s);
  cb = 1;
  for (cb = 1; cb <= bm; cb <<= 1) {
    partner = rank ^ cb;
    if (((rank | (cb - 1)) >= (nump - 1))
        && ((cb - (bm * 2 - nump)) > 0)) {
      scount = sendcount * (cb - (bm * 2 - nump));
    } else {
      scount = sendcount * cb;
    }
    if (((partner | (cb - 1)) >= (nump - 1))
        && ((cb - (bm * 2 - nump)) > 0)) {
      rcount = recvcount * (cb - (bm * 2 - nump));
    } else {
      rcount = recvcount * cb;
    }

    sendind = (rank & (~(cb - 1))) * (recvcount * s);
    recvind = (partner & (~(cb - 1))) * (recvcount * s);
    if (partner >= nump) {      // do a recv from last processor!
      if ((rank != (nump - 1))
          && ((partner & (~(cb - 1))) < nump)) {
        MPI_Recv (recvbuf + recvind, rcount, recvtype,
                  (nump - 1), -14, comm, &status);
      }
      continue;
    }
    MPI_Sendrecv (recvbuf + sendind, scount, sendtype, partner, -14,
                  recvbuf + recvind, rcount, recvtype, partner, -14,
                  comm, &status);
    // Now the last processor serves for the missing ones 
    if (rank == (nump - 1)) {
      for (i = rank + 1; i < bm * 2; i++) {
        partner = i ^ cb;
        if (partner >= rank)
          continue;
        MPI_Send (recvbuf + sendind, scount, sendtype, partner, -14, comm);
      }
    }
  }
  return 0;
}

// tag -19 is used for allgather.
int
MPI_Allgatherv (void *sendbuf, int sendcount, MPI_Datatype sendtype,
                void *recvbuf, int *recvcount, int *displs,
                MPI_Datatype recvtype, MPI_Comm comm)
{
  int nump, rank, s, i;
  int bm, nb, cb;
  int partner, cont;
  int scount, rcount, sendind, recvind;
  MPI_Status status;

  MPI_Comm_size (comm, &nump);
  MPI_Comm_rank (comm, &rank);

  if (rank == -1) {
    printf ("Wrong communicator in calling MPI_Allgather\n");
    fflush (stdout);
    return 1;
  }


  s = _USF_mpi_sizeof (recvtype);
  memcpy (recvbuf + displs[rank], sendbuf, sendcount);
  for (i = 0; i < nump; i++) {
    MPI_Bcast (recvbuf + displs[i], recvcount[i], recvtype, i, comm);
  }
  return MPI_SUCCESS;
}
