// MPI derived datatype implementation
// Created by Chaney 2002
// Uniersity of San Francisco

#include <string.h>
#include "mpi.h"
int DDType_Count = MPI_Primitive_Type;
_USF_MPI_struct_type *struct_head = NULL;

//MPI pack
int
MPI_Pack (void *pack_data, int in_count, MPI_Datatype datatype, void *buffer,
          int buffer_size, int *position, MPI_Comm comm)
{
  memcpy (buffer + *position, pack_data,
          _USF_mpi_sizeof (datatype) * in_count);
  *position += _USF_mpi_sizeof (datatype) * in_count;
}

//MPI unpack
int
MPI_Unpack (void *buffer, int size, int *position, void *unpack_data,
            int count, MPI_Datatype datatype, MPI_Comm comm)
{
  memcpy (unpack_data, buffer + *position,
          _USF_mpi_sizeof (datatype) * count);
  *position += _USF_mpi_sizeof (datatype) * count;
}

//MPI Address
int
MPI_Address (void *location, MPI_Aint * address)
{
  *address = (MPI_Aint) ((char *) location - (char *) MPI_BOTTOM);
  return MPI_SUCCESS;
}

//_USF_DDType_Get_Type
int
_USF_DDType_Get_Type (int idx, _USF_MPI_struct_type ** st)
{
  _USF_MPI_struct_type *ptr;

  if (idx < DDType_Count) {
    ptr = struct_head;
    while (ptr != NULL) {
      if (ptr->idx == idx) {
        break;
      }
      ptr = ptr->next;
    }
    *st = ptr;
    if (*st != NULL) {
      return MPI_SUCCESS;
    }
    return -1;
  }
  return -1;
}

//MPI Type struct
int
MPI_Type_struct (int count, int blocklens[], MPI_Aint indices[],
                 MPI_Datatype old_types[], MPI_Datatype * newtype)
{
  _USF_MPI_struct_type *newstruct;
  int i;

  newstruct = (_USF_MPI_struct_type *) malloc (sizeof (_USF_MPI_struct_type));
  newstruct->count = count;
  newstruct->block_lengths = (int *) malloc (sizeof (int) * count);
  newstruct->disp = (MPI_Aint *) malloc (sizeof (MPI_Aint) * count);
  newstruct->typelist =
    (MPI_Datatype *) malloc (sizeof (MPI_Datatype) * count);
  for (i = 0; i < count; i++) {
    newstruct->block_lengths[i] = blocklens[i];
    newstruct->disp[i] = indices[i];
    newstruct->typelist[i] = old_types[i];
  }
  newstruct->idx = DDType_Count;
  //link new node to linked list
  if (struct_head == NULL) {
    struct_head = newstruct;
    newstruct->next = NULL;
  } else {
    newstruct->next = struct_head;
    struct_head = newstruct;
  }
  *newtype = DDType_Count;
  DDType_Count++;
  return MPI_SUCCESS;
}

//MPI Type vector
int
MPI_Type_vector (int count, int block_length, int stride,
                 MPI_Datatype element_type, MPI_Datatype * new_mpi_t)
{
  _USF_MPI_struct_type *newstruct;
  int i;

  newstruct = (_USF_MPI_struct_type *) malloc (sizeof (_USF_MPI_struct_type));
  newstruct->count = count;
  newstruct->block_lengths = (int *) malloc (sizeof (int) * count);
  newstruct->disp = (MPI_Aint *) malloc (sizeof (MPI_Aint) * count);
  newstruct->typelist =
    (MPI_Datatype *) malloc (sizeof (MPI_Datatype) * count);
  for (i = 0; i < count; i++) {
    newstruct->block_lengths[i] = block_length;
    newstruct->disp[i] = stride * i;
    newstruct->typelist[i] = element_type;
  }
  newstruct->idx = DDType_Count;
  //add new node to linked list
  if (struct_head == NULL) {
    struct_head = newstruct;
    newstruct->next = NULL;
  } else {
    newstruct->next = struct_head;
    struct_head = newstruct;
  }
  *new_mpi_t = DDType_Count;
  DDType_Count++;
  return MPI_SUCCESS;
}

//MPI Type Contiguous
int
MPI_Type_contiguous (int count, MPI_Datatype old_type,
                     MPI_Datatype * new_mpi_t)
{
  _USF_MPI_struct_type *newstruct;
  int i, size;

  newstruct = (_USF_MPI_struct_type *) malloc (sizeof (_USF_MPI_struct_type));
  newstruct->count = 1;
  newstruct->block_lengths = (int *) malloc (sizeof (int));
  newstruct->disp = (MPI_Aint *) malloc (sizeof (MPI_Aint));
  newstruct->typelist = (MPI_Datatype *) malloc (sizeof (MPI_Datatype));
  size = _USF_mpi_sizeof (old_type);
  newstruct->block_lengths[0] = count;
  newstruct->disp[0] = 0;
  newstruct->typelist[0] = old_type;
  newstruct->idx = DDType_Count;
  if (struct_head == NULL) {
    struct_head = newstruct;
    newstruct->next = NULL;
  } else {
    newstruct->next = struct_head;
    struct_head = newstruct;
  }
  *new_mpi_t = DDType_Count;
  DDType_Count++;
  return MPI_SUCCESS;
}

//MPI Type indexed
int
MPI_Type_indexed (int count, int block_lengths[], MPI_Aint displacements[],
                  MPI_Datatype old_type, MPI_Datatype * new_mpi_t)
{
  _USF_MPI_struct_type *newstruct;
  int i;

  newstruct = (_USF_MPI_struct_type *) malloc (sizeof (_USF_MPI_struct_type));
  newstruct->count = count;
  newstruct->block_lengths = (int *) malloc (sizeof (int) * count);
  newstruct->disp = (MPI_Aint *) malloc (sizeof (MPI_Aint) * count);
  newstruct->typelist =
    (MPI_Datatype *) malloc (sizeof (MPI_Datatype) * count);
  for (i = 0; i < count; i++) {
    newstruct->block_lengths[i] = block_lengths[i];
    newstruct->disp[i] = displacements[i];
    newstruct->typelist[i] = old_type;
  }
  newstruct->idx = DDType_Count;
  //add new node to linked list
  if (struct_head == NULL) {
    struct_head = newstruct;
    newstruct->next = NULL;
  } else {
    newstruct->next = struct_head;
    struct_head = newstruct;
  }
  *new_mpi_t = DDType_Count;
  DDType_Count++;
  return MPI_SUCCESS;
}
