#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>
#include <pthread.h>
#include "mpi.h"
#include "gm_thread.h"
#ifdef _USF_MPI_GM_
#include "gm.h"
#endif

//#define DEBUG
#ifdef _USF_MPI_GM_
extern pthread_t _USF_gm_deamon;        /* The gm polling thread thread */
#endif

extern int ack_pending;

int
MPI_Finalize (void)
{
  int i;                        /* Loop variable. */
  int nump;                     /* Number of processors. */
  int rank;                     /* The rank of the current processor. */
  _USF_request_node *curr;      /* The pointer used to traverse the list */
  MPI_Request_Int *req;         /* The pointer used to store MPI_Request_Int */
  _USF_message_header hdr;
  MPI_Status status;

#ifdef _USF_MPI_GM_
  gm_recv_event_t *event;
  while (ack_pending) {
    event = gm_receive (local_port);
    switch (gm_ntoh_u8 (event->recv.type)) {
    case GM_NO_RECV_EVENT:
      break;
    default:
      gm_unknown (local_port, event);
    }
  }

#endif

  MPI_Comm_size (MPI_COMM_WORLD, &nump);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);



#ifdef DEBUG
  printf ("Node %d Finalize start to MPI_wait\n", rank);
  fflush (stdout);

#endif
  /* Wait for the uncompleted MPI_Requests */
  for (curr = _USF_request_hd; curr != NULL; curr = curr->next) {
    /* These requests can't be blocking */
#ifdef DEBUG
    printf ("Node %d Finalize start to _usf_req \n", rank);
    fflush (stdout);

#endif

    req = curr->req;
    if (!(req->done))
      MPI_Wait (&req, &status);
  }
  i = 1;                        /* Used to msg the comm_thread */
#ifdef DEBUG
  printf ("Node %d Finalize start to kill comm_thread\n", rank);
  fflush (stdout);

#endif


#ifdef _USF_MPI_TCP_
  write (_USF_s_pipe[1], &i, sizeof (int));
  pthread_join (_USF_comm_deamon, (void *) &i);
  /* if 0 receive the messages that haven't already sent,
     then send the termination messages. Like a mpi barrier
     if != 0 send the termination message to zero and wait for
     the termination message from zero
   */

#ifdef DEBUG
  printf ("Node %d Finalize start to recv and send\n", rank);
  fflush (stdout);

#endif

  if (rank == 0) {
    for (i = 1; i < nump; i++) {
      if (_USF_finalize_list[i] != 1) {
        _USF_recv_all (i, sizeof (_USF_message_header), &hdr,
                       "Couldn't receive comm term mes");
      }
    }
    hdr.source = 0;
    hdr.dest = i;
    hdr.tag = 0;
    hdr.comm = MPI_COMM_WORLD;
    hdr.size = 0;
    hdr.type = _USF_COMMTERM;
    for (i = 1; i < nump; i++) {
      hdr.dest = i;
      _USF_send_all (i, sizeof (_USF_message_header), &hdr,
                     "Couldn't send comm term mes");
    }
  } else {
    hdr.source = rank;
    hdr.dest = 0;
    hdr.tag = 0;
    hdr.comm = MPI_COMM_WORLD;
    hdr.size = 0;
    hdr.type = _USF_COMMTERM;
    _USF_send_all (0, sizeof (_USF_message_header), &hdr,
                   "Couldn't send comm term mes");
    _USF_recv_all (0, sizeof (_USF_message_header), &hdr,
                   "Couldn't receive comm term mes");
    if (hdr.type != _USF_COMMTERM) {
      printf ("Should have been a comm_term mes\n");
      exit (0);
    }
  }

#endif
  /* if 0 receive the messages that haven't already sent,
     then send the termination messages. Like a mpi barrier
     if != 0 send the termination message to zero and wait for
     the termination message from zero 
   */

  //--killing gm thread
#ifdef DEBUG
  printf ("Node %d Finalize start to kill gm_thread\n", rank);
  fflush (stdout);
#endif

#ifdef _USF_MPI_GM_
  gm_thread_end ();
  pthread_join (_USF_gm_deamon, (void *) &i);
#endif


#ifdef DEBUG
  printf ("Node %d Starting _USF_finalize\n", rank);
  fflush (stdout);
#endif


  _USF_finalize (rank, nump);
  pool_drain (_USF_mpi_request_int_pool);
  pool_drain (_USF_request_node_pool);
  pool_drain (_USF_sysreq_node_pool);
  /* Free the receive buffer  Not implemented yet! */
  return 0;
}
