#include<stdio.h>
#include<stdlib.h>
#include<sys/types.h>
#include<unistd.h>
#include"errno.h"
#include"mpi.h"

// tag -12 is used for gather.
int
MPI_Gather (void *sendbuf, int sendcount, MPI_Datatype sendtype,
            void *recvbuf, int recvcount, MPI_Datatype recvtype,
            int root, MPI_Comm comm)
{
  int nump, rank, i, j, s;
  MPI_Request *req;
  MPI_Status *status;

  MPI_Comm_size (comm, &nump);
  MPI_Comm_rank (comm, &rank);

  s = _USF_mpi_sizeof (recvtype);

  req = (MPI_Request *) malloc ((nump - 1) * sizeof (MPI_Request));
  status = (MPI_Status *) malloc ((nump - 1) * sizeof (MPI_Status));

  if (rank == root) {
    for (i = 0, j = 0; i < nump; i++, j++) {
      if (i == rank) {
        j--;
        continue;
      }
      MPI_Irecv ((recvbuf + (i * recvcount * s)), recvcount,
                 recvtype, i, -12, comm, &(req[j]));
    }

    memcpy ((recvbuf + (root * recvcount * s)), sendbuf, recvcount * s);
    MPI_Waitall ((nump - 1), req, status);
  } else {
    MPI_Send (sendbuf, sendcount, sendtype, root, -12, comm);
  }
  return 0;
}

//tag -17 is used for gatherv
int
MPI_Gatherv (void *sendbuf, int sendcount, MPI_Datatype sendtype,
             void *recvbuf, int *recvcount, int *displs,
             MPI_Datatype recvtype, int root, MPI_Comm comm)
{
  int nump, rank, i, j, s;
  MPI_Request *req;
  MPI_Status *status;

  MPI_Comm_size (comm, &nump);
  MPI_Comm_rank (comm, &rank);

  if (rank == -1) {
    printf ("Wrong communicator in calling MPI_Gather\n");
    fflush (stdout);
    return 1;
  }

  s = _USF_mpi_sizeof (recvtype);

  req = (MPI_Request *) malloc ((nump - 1) * sizeof (MPI_Request));
  status = (MPI_Status *) malloc ((nump - 1) * sizeof (MPI_Status));

  if (rank == root) {
    for (i = 0, j = 0; i < nump; i++, j++) {
      if (i == rank) {
        j--;
        continue;
      }
      MPI_Irecv (recvbuf + displs[i], recvcount[i],
                 recvtype, i, -17, comm, &(req[j]));
    }

    memcpy (recvbuf + displs[root], sendbuf, recvcount[root] * s);
    MPI_Waitall ((nump - 1), req, status);
  } else {
    MPI_Send (sendbuf, sendcount, sendtype, root, -17, comm);
  }
  return 0;
}
