// The mrun program is the equivalent of mpirun, it will initiate the
// calls to the processes.
// Currently only the number of processes and the name of the runnable 
// file is going to be implemented.

#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include"debug.h"
#include"mpi.h"

// The arguments that will be used when calling the main function of
// the processes:
// 1 - The rank of the processor.
// 2 - The number of processors.
// 3 - The address of the current processor.

int
main (int argc, char **argv)
{

  char *cmd_str;                // Commandline string.
  int tmp;                      // The temporary variable.
  int i, j;                     // Loop variable.
  int nump = 1;                 // the number of processors.
  char *nump_str;
  int pos_called_file = 1;      // the pos of file name in argv[], default is 1
  char *cur_num;                // Used as a parameter to itostr function.
  char *my_name;                // The hostname that runs the program.
  pid_t *children;              // The pids of the child processes.
  int err;                      // The error value.
  char cwdbuf[256];             // Buffer for current working directory


  cur_num = (char *) malloc (8 * sizeof (char));
  cmd_str = (char *) malloc (256 * sizeof (char));
  my_name = (char *) malloc (128 * sizeof (char));

  // get the num of processors
  for (i = 1; i < argc; i++) {
    if (strcmp (argv[i], "-np") == 0) {
      pos_called_file = i + 2;
      nump = atoi (argv[i + 1]);
      nump_str = argv[i + 1];
    } else
      break;
  }

  // get the host name
  if (gethostname (my_name, 30) != 0)
    return Error (ER_HOSTNM);

  // call nump copies of programs  
  if (nump == 1) {              // run with only one processor
    if ((err = _USF_make_conn_list (my_name, 1)) != 0)
      return err;
    // Start the program on processors
    memset (cmd_str, '\0', 80);
    if (strcmp (my_name, _USF_conn_list[0].hostname) != 0) {
      strcat (cmd_str, "rsh ");
      strcat (cmd_str, _USF_conn_list[0].hostname);
    }
    if (argv[pos_called_file][0] == '/') {      // A complete filename
      strcat (cmd_str, " ");
      strcat (cmd_str, argv[pos_called_file]);  // get the called file name
    } else {                    // Need to concat this current directory
      strcat (cmd_str, " ");
      strcat (cmd_str, getcwd (cwdbuf, 256));
      strcat (cmd_str, "/");
      strcat (cmd_str, argv[pos_called_file]);  // get the called file name
    }
    strcat (cmd_str, " 0 1 ");  // rank, number of procs
    strcat (cmd_str, my_name);  // hostname

    // get the parameters for called program
    for (j = pos_called_file + 1; j < argc; j++) {
      strcat (cmd_str, " ");
      strcat (cmd_str, argv[j]);
    }

    nump = 1;
    children = (pid_t *) malloc (nump * sizeof (pid_t));
    children[0] = fork ();
    if (!children[0]) {         //The child process
      if (system (cmd_str) != 0) {
        return Error (ER_FILENAME);
      }
      exit (0);
    }
  } else if (nump > 1) {        // specify the number of processors
    if ((err = _USF_make_conn_list (my_name, nump)) != 0)
      return err;

    children = (pid_t *) malloc (nump * sizeof (pid_t));
    for (i = 0; i < nump; i++) {        // Start the program on processors
      memset (cmd_str, '\0', 80);
      if (strcmp (my_name, _USF_conn_list[i].hostname) != 0) {
        strcat (cmd_str, "rsh ");
        strcat (cmd_str, _USF_conn_list[i].hostname);
      }

      if (argv[pos_called_file][0] == '/') {    // A complete filename
        strcat (cmd_str, " ");
        strcat (cmd_str, argv[pos_called_file]);        // get the called file name
      } else {                  // Need to concat this current directory
        strcat (cmd_str, " ");
        strcat (cmd_str, getcwd (cwdbuf, 256));
        strcat (cmd_str, "/");
        strcat (cmd_str, argv[pos_called_file]);        // get the called file name
      }

      strcat (cmd_str, " ");
      sprintf (cur_num, "%d", i);       // rank
      strcat (cmd_str, cur_num);
      strcat (cmd_str, " ");
      strcat (cmd_str, nump_str);       // number of procs
      strcat (cmd_str, " ");
      strcat (cmd_str, my_name);        // hostname

      for (j = pos_called_file + 1; j < argc; j++) {
        strcat (cmd_str, " ");
        strcat (cmd_str, argv[j]);
      }

      children[i] = fork ();
      if (!children[i]) {       //The child process
        DBG (("here: %s\n", cmd_str));
        if (system (cmd_str) != 0) {
          return Error (ER_FILENAME);
        }
        exit (0);
      }
    }
  } else {                      // Error
    return Error (ER_INCORRECT);
  }
  for (i = 0; i < nump; i++) {  // wait for the children
    waitpid (children[i], NULL, 0);
  }


  return 0;
}
