#include "mpi.h"
#include "sys/time.h"

int
MPI_Type_size (MPI_Datatype datatype, int *size)
{

  *size = _USF_mpi_sizeof (datatype);

  return (MPI_SUCCESS);
}



double
MPI_Wtime ()
{
  double t1;
  struct timeval t2;
  gettimeofday (&t2, NULL);
  t1 = t2.tv_sec + t2.tv_usec / 1000000.0;
  return t1;
}


int
MPI_Error_string (int errorcode, char *string, int *resultlen)
{
  strcpy (string, "not implemented");

}


int
MPI_Reduce_scatter (void *sendbuf, void *recvbuf, int *recvcnts,
                    MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
{


}

int
MPI_Abort (MPI_Comm comm, int errorcode)
{

  return MPI_ERR_UNKNOWN;
}


int
MPI_Comm_group (MPI_Comm comm, MPI_Group * group)
{
  int i;
  for (i = 0; i < _USF_num_of_comm; i++) {
    if (_USF_comm_list[i].comm == comm) {
      *group = _USF_comm_list[i].group;
      return 0;
    }
  }
  return -1;
}

int
MPI_Group_incl (MPI_Group old_group,    // in
                int new_group_size,     // in
                int ranks_in_old_group[],       // in
                MPI_Group * new_group   // out
  )
{
  int i, j;
  // More code needs to be added for groups more than 16
  if (_USF_group_max == _USF_num_of_group)
    return -1;
  *new_group = _USF_group_ref;
  _USF_group_list[_USF_num_of_group].group_no_old = old_group;
  _USF_group_list[_USF_num_of_group].group_no_local = _USF_group_ref;
  _USF_group_list[_USF_num_of_group].group_size = new_group_size;

  _USF_get_group_num (old_group, &j);


  for (i = 0; i < new_group_size; i++) {
    // set the rank in old group
    _USF_group_list[_USF_num_of_group].rank_old[i] = ranks_in_old_group[i];
    // set the rank in global group 0
    _USF_group_list[_USF_num_of_group].rank_global[i] =
      _USF_group_list[j].rank_global[ranks_in_old_group[i]];
    // set the rank in MPI_COMM_WORLD


  }

  _USF_group_ref++;             // increase the group reference number
  _USF_num_of_group++;          //
  return 0;
}

// find coressponding position in group_list
int
_USF_get_group_num (MPI_Group group, int *group_num)
{
  int i;
  for (i = 0; i < _USF_num_of_group; i++) {
    if (_USF_group_list[i].group_no_local == group) {
      *group_num = i;
      return 0;
    }
  }

  return 1;
}


int
MPI_Comm_create (MPI_Comm old_comm,     // in
                 MPI_Group new_group,   // in
                 MPI_Comm * new_comm    // out
  )
{
  int i, j, k, rank;
  // More code needs to be added for communicators more than 16
  if (_USF_comm_max == _USF_num_of_comm)
    return -1;
  *new_comm = _USF_comm_ref;

  MPI_Comm_rank (old_comm, &rank);

  _USF_get_group_num (new_group, &j);
  _USF_get_comm_num (old_comm, &k);

  _USF_comm_list[_USF_num_of_comm].group = new_group;
  _USF_comm_list[_USF_num_of_comm].comm = _USF_comm_ref;
  _USF_comm_list[_USF_num_of_comm].old_comm = old_comm;
  _USF_comm_list[_USF_num_of_comm].nump = _USF_group_list[j].group_size;

  _USF_comm_list[_USF_num_of_comm].rank = -1;
  // -1 means not belonging to this communicator

  for (i = 0; i < _USF_comm_list[_USF_num_of_comm].nump; i++) {
    // rank is the processor's rank in old communicator
    if (rank == _USF_group_list[j].rank_old[i])
      _USF_comm_list[_USF_num_of_comm].rank = i;

  }

  _USF_comm_ref++;
  _USF_num_of_comm++;

  return 0;
}

int
_USF_get_comm_num (MPI_Comm comm, int *comm_num)
{
  int i;
  for (i = 0; i < _USF_num_of_comm; i++) {
    if (_USF_comm_list[i].comm == comm) {
      *comm_num = i;
      return 0;
    }
  }
  *comm_num = -1;
  return 1;
}

int
MPI_Group_translate_ranks (MPI_Group group_a, int n, int *ranks_a,
                           MPI_Group group_b, int *ranks_b)
{
  return 0;
}


int
MPI_Comm_free (MPI_Comm * commp)
{
  return 0;
}


int
MPI_Comm_split (MPI_Comm comm, int color, int key, MPI_Comm * comm_out)
{
  int i, j, k, m, i1, i2;
  int output = 0;
  int comm_num;
  int size, rank;
  int *buf, *workingarray;
  int local[2];
  int tempColor;
  MPI_Comm tempComm;
  int *group_rank = NULL;
  MPI_Group old_group, new_group;

  MPI_Comm_size (comm, &size);
  MPI_Comm_rank (comm, &rank);

  _USF_get_comm_num (comm, &comm_num);

  MPI_Comm_group (comm, &old_group);

  buf = (int *) malloc (size * 2 * sizeof (int));
  workingarray = (int *) malloc (size * sizeof (int));

  if (buf == NULL) {
    printf ("Memory Error\n");
    return;
  }

  for (i = 0; i < size; i++) {
    workingarray[i] = 0;
  }

  local[0] = color;
  local[1] = key;

  group_rank = (int *) malloc (size * sizeof (int));

  MPI_Allgather (local, 2, MPI_INT, buf, 2, MPI_INT, comm);



  for (i = 0; i < size; i++) {
    if (workingarray[i] == 1)
      continue;
    tempColor = buf[2 * i];
    k = 0;
    for (j = i; j < size; j++) {
      if (tempColor == buf[j * 2] && workingarray[j] == 0) {
        group_rank[k] = j;
        k++;
        workingarray[j] = 1;
        if (j == rank)
          output = 1;
      }
    }

    for (i1 = 0; i1 < k - 1; i1++) {
      for (i2 = i1; i2 < k - 1; i2++)
        if (buf[group_rank[i2] * 2 + 1] > buf[group_rank[i2 + 1] * 2 + 1]) {
          m = group_rank[i2 + 1];
          group_rank[i2 + 1] = group_rank[i2];
          group_rank[i2] = m;
        }
    }

    MPI_Group_incl (old_group, k, group_rank, &new_group);
    if (output) {
      MPI_Comm_create (comm, new_group, comm_out);
      output = 0;
    } else
      MPI_Comm_create (comm, new_group, &tempComm);
  }

  if (color == MPI_UNDEFINED) {
    *comm_out = MPI_COMM_NULL;
  }

  return MPI_SUCCESS;

}



// check the communicator

int
_USF_check_comm (MPI_Comm comm, int *target)
{
  int rank, i;

  if (*target == MPI_ANY_SOURCE)
    return 0;

  MPI_Comm_rank (comm, &rank);
  if (rank == -1)
    return 1;
  else {
    _USF_get_comm_num (comm, &i);

    _USF_get_group_num (_USF_comm_list[i].group, &i);

    *target = _USF_group_list[i].rank_global[*target];

    return 0;
  }


}

int
MPI_Group_rank (MPI_Group group, int *rank)
{
  int group_num, global_rank, i;

  *rank = -1;
  if (_USF_get_comm_num (group, &group_num))
    return 1;
  MPI_Comm_rank (MPI_COMM_WORLD, &global_rank);

  for (i = 0; i < _USF_group_list[group_num].group_size; i++)
    if (_USF_group_list[group_num].rank_global[i] == global_rank)
      *rank = i;

  if (*rank == -1)
    return 1;
  return 0;

}


int
MPI_Group_size (MPI_Group group, int *size)
{
  int group_num;

  *size = -1;
  if (_USF_get_comm_num (group, &group_num))
    return 1;

  *size = _USF_group_list[group_num].group_size;

}
