#include<stdio.h>
#include<stdlib.h>
#include<time.h>
#include"mpi.h"
#include"debug.h"

extern int thread_recving;

int
MPI_Probe (int source, int tag, MPI_Comm comm, MPI_Status * status)
{
  int flag, rank, nump;         /* Flag value used for Iprobe */
  MPI_Request_Int *req;         /* The pointer used to init the MPI Request */
  _USF_request_node *node;      /* The pointer used to store
                                 * the MPI Request in.*/

  node = (_USF_request_node *) pool_take (_USF_request_node_pool);
  req = (MPI_Request_Int *) pool_take (_USF_mpi_request_int_pool);
  req->active = 1;
  req->done = 0;
  req->end_of_transmission = 0;
  req->type = _USF_PROBE;
  req->blocking = 1;
  req->posted = 0;
  req->tag = MPI_ANY_TAG;
  req->peer = source;
  req->comm = comm;
  req->multiple = 0;
  req->nchunk = 0;
  req->node = node;
  node->req = req;

  MPI_Comm_size (MPI_COMM_WORLD, &nump);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);


  pthread_mutex_lock (&_USF_mut_req);
  _USF_request_inshd (node);
  pthread_mutex_unlock (&_USF_mut_req);

#ifdef DEBUG
  printf ("rank %d Probe start\n", _USF_comm_list[0].rank);
  fflush (stdout);
#endif
  flag = 0;
  while (!flag) {
    MPI_Iprobe (source, tag, comm, &flag, status);
    if (!flag) {
#ifndef SINGLE_THREAD
      pthread_mutex_lock (&_USF_mut_block);
      if (req->end_of_transmission == 0)
        pthread_cond_wait (&_USF_cond_block, &_USF_mut_block);
      pthread_mutex_unlock (&_USF_mut_block);
#else

#ifdef _USF_MPI_TCP_
      int i;                    /* Loop variable */

      fd_set set;               /* the set that is used for select */
      int max;                  /* the maximum file descriptor */
      int finished = 0;         /* The loop conditional */
      int nu;




      while (!(req->end_of_transmission)) {

        max = _USF_TCP_prepare_set (&set);
        if ((nu = pselect (max + 1, &set, NULL, NULL, NULL, NULL)) == -1)
          Error (ERF_SELECT);
        DBG (("rank %d, woken up from select! sel %d \n", rank, nu));
        for (i = 0; i < nump; i++) {
          if (i == rank)
            continue;
          if (FD_ISSET (_USF_conn_list[i].sd, &set)) {
            DBG (("rank %d handling message %d start\n", rank, i));
            _USF_handle_msg (i);
            DBG (("rank %d handling message %d end\n", rank, i));
          }
        }
      }
#endif

#ifdef _USF_MPI_GM_
      while (!(req->end_of_transmission)) {
        thread_recving = 1;
        gm_thread ();
      }
#endif

#endif
    }

  }
  pthread_mutex_lock (&_USF_mut_req);
  _USF_request_remove (node);
  pthread_mutex_unlock (&_USF_mut_req);
#ifdef DEBUG
  printf ("rank %d Probe end\n", _USF_comm_list[0].rank);
  fflush (stdout);
#endif
  return 0;
}
