#include<stdio.h>
#include<stdlib.h>
#include<sys/types.h>
#include<unistd.h>
#include<pthread.h>
#include"errno.h"
#include"mpi.h"

#define LP printf

// tag -13 is used for gather.
int
MPI_Scatter (void *sendbuf, int sendcount, MPI_Datatype sendtype,
             void *recvbuf, int recvcount, MPI_Datatype recvtype,
             int root, MPI_Comm comm)
{
  int nump, rank, i, j, s;
  int dene[5];
  MPI_Request *req;
  MPI_Status *status;

  MPI_Comm_size (comm, &nump);
  MPI_Comm_rank (comm, &rank);

  if (rank == -1) {
    printf ("Wrong communicator in calling MPI_Scatter\n");
    fflush (stdout);
    return 1;
  }

  s = _USF_mpi_sizeof (recvtype);

  req = (MPI_Request *) malloc ((nump - 1) * sizeof (MPI_Request));
  status = (MPI_Status *) malloc ((nump - 1) * sizeof (MPI_Status));

  if (rank == root) {
    for (i = 0, j = 0; i < nump; i++, j++) {
      if (i == rank) {
        j--;
        continue;
      }
      MPI_Isend ((sendbuf + (i * sendcount * s)), sendcount,
                 sendtype, i, -13, comm, &(req[j]));
    }

    memcpy (recvbuf, (sendbuf + (root * sendcount * s)), sendcount * s);
    MPI_Waitall ((nump - 1), req, status);
  } else {
    MPI_Recv (recvbuf, recvcount, recvtype, root, -13, comm, status);
  }
  return 0;
}

// tag -18 is used for gather.
int
MPI_Scatterv (void *sendbuf, int *sendcount, int *displs,
              MPI_Datatype sendtype, void *recvbuf, int recvcount,
              MPI_Datatype recvtype, int root, MPI_Comm comm)
{
  int nump, rank, i, j, s;
  int dene[5];
  MPI_Request *req;
  MPI_Status *status;

  MPI_Comm_size (comm, &nump);
  MPI_Comm_rank (comm, &rank);

  if (rank == -1) {
    printf ("Wrong communicator in calling MPI_Scatter\n");
    fflush (stdout);
    return 1;
  }

  s = _USF_mpi_sizeof (recvtype);

  req = (MPI_Request *) malloc ((nump - 1) * sizeof (MPI_Request));
  status = (MPI_Status *) malloc ((nump - 1) * sizeof (MPI_Status));

  if (rank == root) {
    for (i = 0, j = 0; i < nump; i++, j++) {
      if (i == rank) {
        j--;
        continue;
      }
      MPI_Isend (sendbuf + displs[i], sendcount[i],
                 sendtype, i, -18, comm, &(req[j]));
    }

    memcpy (recvbuf, sendbuf + displs[root], sendcount[root] * s);
    MPI_Waitall ((nump - 1), req, status);
  } else {
    MPI_Recv (recvbuf, recvcount, recvtype, root, -18, comm, status);
  }
  return 0;
}
