#include<stdio.h>
#include"mpi.h"

_USF_sysreq_node *_USF_sysreq_hashtable[_USF_MAXPROCESSORS];

_USF_sysreq_node *
_USF_sysreq_lookup (int tag, int source, int type, int comm)
{
  int h = 0;
  int i = 0;
  int nump = 0;
  _USF_sysreq_node *node = NULL;

  MPI_Comm_size (comm, &nump);  // Modified by Qing Huang

  if (type == _USF_SEND) {      // Search for Send request

    if (source == MPI_ANY_SOURCE) {
      for (i = 0; i < nump; i++) {
        h = i % _USF_MAXPROCESSORS;
        for (node = _USF_sysreq_hashtable[h]; node != NULL; node = node->next) {
#ifdef DEBUG
          printf (" ---> s %d sr %d ty %d c %d\n", tag, i, type, comm);
          printf (" ---> s %d sr %d ty %d c %d\n", node->tag, node->source,
                  node->type, node->comm);
#endif
          if (((tag == node->tag) ||
               ((tag == MPI_ANY_TAG) && (tag >= -1)))
              && (i == node->source)
              && (type == node->type)
              && (comm == node->comm)) {
#ifdef DEBUG
            printf (" ---> Found Send Sys node and return\n");
            fflush (stdout);
#endif
            return node;
          }
        }
      }
#ifdef DEBUG
      printf (" ---> Not Found Send Sys node and return\n");
      fflush (stdout);
#endif
      return node;
    }
    h = source % _USF_MAXPROCESSORS;
    for (node = _USF_sysreq_hashtable[h]; node != NULL; node = node->next) {
      if (((tag == node->tag) || (tag == MPI_ANY_TAG))  
	  // Modified by Qing Huang
          && (source == node->source)
          && (type == node->type) && (comm == node->comm)) {
        return node;
      }
    }
  } else if (type == _USF_RECEIVE) {    // Search for Recieve request

    h = source % _USF_MAXPROCESSORS;
    for (node = _USF_sysreq_hashtable[h]; node != NULL; node = node->next) {
      if (((tag == node->tag) || (node->tag == MPI_ANY_TAG))    
	  // Modified by Qing Huang
          && (source == node->source)
          && (type == node->type) && (comm == node->comm)) {
        return node;
      }
    }

  }


  return node;
}


_USF_sysreq_node *
_USF_sysreq_remove (int tag, int source, int type, int comm)
{
  int h = 0;
  int i = 0;
  int nump = 0;
  _USF_sysreq_node *node = NULL, *prev = NULL;

  MPI_Comm_size (comm, &nump);

  if (type == _USF_SEND) {
#ifdef DEBUG

    printf (" --->%d Start to remove type _USF_SEND\n", nump);
    fflush (stdout);
#endif

    if (source == MPI_ANY_SOURCE) {
      for (i = 0; i < nump; i++) {
        h = i % _USF_MAXPROCESSORS;
        for (prev = node = _USF_sysreq_hashtable[h]; node != NULL;
             prev = node, node = node->next) {
#ifdef DEBUG
          printf (" ---> r %d sr %d ty %d c %d\n", tag, i, type, comm);
          printf (" ---> r %d sr %d ty %d c %d\n", node->tag, node->source,
                  node->type, node->comm);
#endif
          if (((tag == node->tag) || ((tag == MPI_ANY_TAG) && (tag >= -1)))
	      // Modified by Qing Huang
              /*may need || (node->tag == -1)) MPI_ANY_TAG */
              && (i == node->source)    // Modified by Qing Huang 
              && (type == node->type)
              && (comm == node->comm)) {
#ifdef DEBUG
            printf (" ---> Find Send Sys node and will remove it\n");
            fflush (stdout);
#endif
            if (node == prev) {
              _USF_sysreq_hashtable[h] = node->next;
            } else {
              prev->next = node->next;
            }
#ifdef DEBUG
            printf (" ---> Found Send Sys node and removed\n");
            fflush (stdout);
#endif
            return node;
          }
        }
      }
#ifdef DEBUG
      printf (" ---> Not Found Send Sys node and not removed\n");
      fflush (stdout);
#endif
      return node;
    }
    h = source % _USF_MAXPROCESSORS;
    for (prev = node = _USF_sysreq_hashtable[h];
         node != NULL; prev = node, node = node->next) {
      if (((tag == node->tag) || (tag == MPI_ANY_TAG))  
	  // Modified by Qing Huang
          /*may need || (node->tag == -1)) MPI_ANY_TAG */
          && ((source == node->source) || (source == MPI_ANY_SOURCE))   
	  // Modified by Qing Huang 
          && (type == node->type) && (comm == node->comm)) {
        if (node == prev) {
          _USF_sysreq_hashtable[h] = node->next;
        } else {
          prev->next = node->next;
        }
#ifdef DEBUG
        printf (" ---> Found Send Sys node and removed\n");
        fflush (stdout);
#endif
        break;
      }
    }
  }

  else if (type == _USF_RECEIVE) {

#ifdef DEBUG
    printf (" ---> Start to remove type _USF_RECEIVE\n");
    fflush (stdout);
#endif

    h = source % _USF_MAXPROCESSORS;
    for (prev = node = _USF_sysreq_hashtable[h]; node != NULL;
         prev = node, node = node->next) {
      if (((tag == node->tag) || (node->tag == MPI_ANY_TAG) || tag == MPI_ANY_TAG)      
	  // Modified by Qing Huang
          /*may need || (node->tag == -1)) MPI_ANY_TAG */
          && (source == node->source)   // Modified by Qing Huang 
          && (type == node->type) && (comm == node->comm)) {
        if (node == prev) {
          _USF_sysreq_hashtable[h] = node->next;
        } else {
          prev->next = node->next;
        }
#ifdef DEBUG
        printf (" ---> Found Recv Sys node and removed\n");
        fflush (stdout);
#endif
        break;
      }
    }

  }


  return node;
}

#ifdef _USF_MPI_TCP_
void
_USF_sysreq_insert (int tag, int source, int type, int comm, int size,
                    _USF_sysreq_node * value)
{
  int h;
  _USF_sysreq_node *node = NULL, *prev = NULL;

  h = source % _USF_MAXPROCESSORS;
  value->next = NULL;
  value->tag = tag;
  value->source = source;
  value->type = type;
  value->comm = comm;
  value->size = size;

  if (_USF_sysreq_hashtable[h] == NULL) {
    _USF_sysreq_hashtable[h] = value;
    return;
  }
  for (node = _USF_sysreq_hashtable[h];
       node != NULL; prev = node, node = node->next);
  prev->next = value;
  return;
}
#endif


#ifdef _USF_MPI_GM_
void
_USF_sysreq_insert (int tag, int source, int type, int comm, int size,
                    void *in_buf, _USF_sysreq_node * value)
{
  int h;
  _USF_sysreq_node *node = NULL, *prev = NULL;

  h = source % _USF_MAXPROCESSORS;
  value->next = NULL;
  value->tag = tag;
  value->source = source;
  value->type = type;
  value->comm = comm;
  value->size = size;
  value->in_buf = in_buf;
  if (_USF_sysreq_hashtable[h] == NULL) {
    _USF_sysreq_hashtable[h] = value;
    return;
  }
  for (node = _USF_sysreq_hashtable[h];
       node != NULL; prev = node, node = node->next);
  prev->next = value;
  return;
}
#endif

void
_USF_sysreq_free ()
{
  int i;
  _USF_sysreq_node *curr;

  for (i = 0; i < _USF_MAXPROCESSORS; i++) {
    while (_USF_sysreq_hashtable[i] != NULL) {
      curr = _USF_sysreq_hashtable[i];
      _USF_sysreq_hashtable[i] = _USF_sysreq_hashtable[i]->next;
      if (curr != NULL)
        free (curr);
    }
  }
  return;
}
