//  Error message function
//  Get the size of MPI datatype
//  
//  make connection list and set initial values
//


#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<errno.h>
#include<string.h>
#include"debug.h"
#include"mpi.h"
#include "gm.h"

_USF_conn_struct *_USF_conn_list;

#define DEBUG

int
Error (int wh)
{
  char *str;
  int rank;

  str = (char *) malloc (200 * sizeof (char));

  MPI_Comm_rank (MPI_COMM_WORLD, &rank);


  switch (wh) {
  case ER_INCORRECT:
    strcat (str, "The correct usage:\n");
    strcat (str, "mrun <execname> or mrun -np <#proc> <execname>\n");
    break;
  case ER_FILENAME:
    strcat (str, "Filename missing, or not executable.\n");
    break;
  case ER_NUMP:
    strcat (str, "The number of procs should be greater than zero.\n");
    break;
  case ER_NEMACS:
    strcat (str, "There aren't enough number of processors.\n");
    break;
  case ER_HOSTNM:
    strcat (str, "The host name unavailable.\n");
    break;
  case ER_MACFILE:
    strcat (str, "The machines file is missing.\n");
    break;
  case ERF_SOCKETINIT:
    strcat (str, "Unable to initialize socket.\n");
    break;
  case ERF_BINDINIT:
    strcat (str, "Unable to bind socket.\n");
    break;
  case ERF_ACCEPT:
    strcat (str, "Unable to accept new connection.\n");
    break;
  case ERF_RECV:
    strcat (str, "Unable to receive.\n");
    break;
  case ERF_SEND:
    strcat (str, "Unable to send.\n");
    break;
  case ERF_GETHE:
    strcat (str, "Unable to get address by hostname.\n");
    break;
  case ERF_CONNECT:
    strcat (str, "Unable to connect.\n");
    break;
  case ERF_SETSOCKOPT:
    strcat (str, "Setsockopt error.\n");
    break;
  case ERF_LISTEN:
    strcat (str, "Unable to listen.\n");
    break;
  case ERF_SELECT:
    strcat (str, "Select error.\n");
    break;
  case ERF_PTHREAD:
    strcat (str, "Pthread creation error.\n");
    break;
  default:
    strcat (str, "Unexpected error.\n");
    break;
  }
  printf ("Error num--> %d : %s", wh, str);
  system ("hostname");
  fflush (stdout);
  free (str);
  exit (0);                     /* may change to only exit if fatal error */
  return wh;
}

int
_USF_mpi_sizeof (MPI_Datatype ty)
{

  switch (ty) {
  case MPI_CHAR:
    return sizeof (char);
    break;
  case MPI_SHORT:
    return sizeof (short);
    break;
  case MPI_INT:
    return sizeof (int);
    break;
  case MPI_LONG:
    return sizeof (long);
    break;
  case MPI_UNSIGNED_CHAR:
    return sizeof (unsigned char);
    break;
  case MPI_UNSIGNED_SHORT:
    return sizeof (unsigned short);
    break;
  case MPI_UNSIGNED:
    return sizeof (unsigned);
    break;
  case MPI_UNSIGNED_LONG:
    return sizeof (unsigned long);
    break;
  case MPI_FLOAT:
    return sizeof (float);
    break;
  case MPI_DOUBLE:
    return sizeof (double);
    break;
  case MPI_LONG_DOUBLE:
    return sizeof (long double);
    break;
  case MPI_BYTE:
    return sizeof (char);
    break;
  case MPI_LONG_LONG:
    return sizeof (long long);
  case MPI_PACKED:
    return sizeof (char);
    break;
  case MPI_LB:
    return 1;
    break;
  case MPI_UB:
    return 1;
    break;
  case MPI_FLOAT_INT:
    return sizeof (float) + sizeof (int);
    break;
  case MPI_DOUBLE_INT:
    return sizeof (double) + sizeof (int);
    break;
  case MPI_LONG_INT:
    return sizeof (long) + sizeof (int);
    break;
  case MPI_SHORT_INT:
    return sizeof (short) + sizeof (int);
    break;
  case MPI_2INT:
    return sizeof (int) + sizeof (int);
    break;
  case MPI_LONG_DOUBLE_INT:
    return sizeof (long double) + sizeof (int);
    break;

  default:
    return -1;
  }
}

int
_USF_make_conn_list (char *my_name, int nump)
{
  FILE *macs;                   // The machines file that mpi will use.
  char *mac_nm;                 // The name of a machine.
  char *cur_dir;                // The current directory.
  int mac_size;                 // The number of processors read.
  char *colonplace;             // The pointer to : in mac_nm str.
  int psize;                    // The number of procs on current ip.
  int i;                        // The loop variable.
  gm_status_t my_status;

  struct gm_port *my_port;

  cur_dir = (char *) malloc (256 * sizeof (char));
  mac_size = 0;
#ifdef _USF_MPI_TCP_
  int default_port = _USF_SERVPORT;
#endif

#ifdef _USF_MPI_GM_
  int default_port = 4;
#endif

  strcpy (cur_dir, getenv ("HOME"));
  strcat (cur_dir, "/machines.kcg");
  macs = fopen (cur_dir, "r");
  if (macs == NULL) {
    printf ("Tryed alternate location, still couldn't find machines file\n");
    return Error (ER_MACFILE);
  }
  _USF_conn_list = (_USF_conn_struct *)
    malloc (nump * sizeof (_USF_conn_struct));
  mac_nm = (char *) malloc (80 * sizeof (char));
  for (; ((fscanf (macs, "%s", mac_nm)) != EOF) && (mac_size < nump);) {
    if ((colonplace = strchr (mac_nm, (int) ':')) == NULL) {
      int i;
      _USF_conn_list[mac_size].hostname = mac_nm;
      for (i = mac_size; i > 0; i--)
        if (!strcmp
            (_USF_conn_list[i - 1].hostname,
             _USF_conn_list[mac_size].hostname))
          break;
      if (i != 0) {
#ifdef _USF_MPI_TCP_
        _USF_conn_list[mac_size].port = _USF_conn_list[i - 1].port + 1;
#endif
#ifdef _USF_MPI_GM_
        _USF_conn_list[mac_size].port_id = _USF_conn_list[i - 1].port_id + 1;
#endif
      } else {
#ifdef _USF_MPI_TCP_
        _USF_conn_list[mac_size].port = default_port;
#endif
#ifdef _USF_MPI_GM_
        _USF_conn_list[mac_size].port_id = default_port;
#endif
      }
      DBG (("%s, %d \n", _USF_conn_list[mac_size].hostname,
            _USF_conn_list[mac_size].port));
      mac_size++;
    } else {
      psize = atoi (colonplace + 1);
      for (i = 0; (i < psize) && (mac_size < nump); i++) {
        *colonplace = '\0';
        _USF_conn_list[mac_size].hostname = mac_nm;
        _USF_conn_list[mac_size].port = _USF_SERVPORT + i;
        DBG (("%s, %d\n", _USF_conn_list[mac_size].hostname,
              _USF_conn_list[mac_size].port));
        mac_size++;
      }
    }
    mac_nm = (char *) malloc (80 * sizeof (char));
  }
  free (mac_nm);
  fclose (macs);
  DBG (("--> util.c out of make\n"));


  for (i = 0; i < nump; i++) {
    _USF_conn_list[i].next_msg = _USF_HEADER_TYPE;
  }
  return 0;
}
