#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include"mpi.h"
#include "gm.h"
#include "debug.h"

extern int debug_counter;
extern int thread_recving;

int
MPI_Wait (MPI_Request * request, MPI_Status * status)
{

  int nump;                     /* # of procs in COMM_WORLD */
  int rank;                     /* the rank in COMM_WORLD */

  MPI_Comm_size (MPI_COMM_WORLD, &nump);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);

  pthread_mutex_lock (&_USF_mut_req);
  status->count = 0;
  if (*request == MPI_REQUEST_NULL) {
    pthread_mutex_unlock (&_USF_mut_req);
    return 1;
  }
  if (!(*request)->active) {
    pthread_mutex_unlock (&_USF_mut_req);
    return 1;
  }

  pthread_mutex_unlock (&_USF_mut_req);


// multiple threaded mode
#ifndef SINGLE_THREAD

  pthread_mutex_lock (&_USF_mut_block);
#ifdef DEBUG
  printf ("rank %d !!Wait!! cond_bl start\n", _USF_comm_list[0].rank);
  fflush (stdout);
#endif
  if ((*request)->end_of_transmission == 0) {
#ifdef DEBUG
    printf ("rank %d waiting ############## %p debug_counter %d\n", rank,
            (*request), debug_counter);
    fflush (stdout);
    printf ("rank %d dest %d type %d tag %d comm %d size %d req %p \n", rank,
            (*request)->peer, (*request)->type, (*request)->tag,
            (*request)->comm, (*request)->size, (*request));
    fflush (stdout);
#endif
    (*request)->blocking = 1;
    pthread_cond_wait (&_USF_cond_block, &_USF_mut_block);

  }
#ifdef DEBUG
  printf ("rank %d !!Wait!! cond_bl end\n", _USF_comm_list[0].rank);
  fflush (stdout);
#endif
  pthread_mutex_unlock (&_USF_mut_block);

// single threaded mode
#else

#ifdef _USF_MPI_TCP_
  int i;                        /* Loop variable */

  fd_set set;                   /* the set that is used for select */
  int max;                      /* the maximum file descriptor */
  int finished = 0;             /* The loop conditional */
  int nu;




  while (!((*request)->end_of_transmission)) {

    max = _USF_TCP_prepare_set (&set);
    if ((nu = pselect (max + 1, &set, NULL, NULL, NULL, NULL)) == -1)
      Error (ERF_SELECT);
    DBG (("rank %d, woken up from select! sel %d \n", rank, nu));
    for (i = 0; i < nump; i++) {
      if (i == rank)
        continue;
      if (FD_ISSET (_USF_conn_list[i].sd, &set)) {
        DBG (("rank %d handling message %d start\n", rank, i));
        _USF_handle_msg (i);
        DBG (("rank %d handling message %d end\n", rank, i));
      }
    }
  }
#endif

#ifdef _USF_MPI_GM_
  while (!((*request)->end_of_transmission)) {
    thread_recving = 1;
    gm_thread ();
  }
#endif

#endif




  /* The message has arrived */
  pthread_mutex_lock (&_USF_mut_req);
  status->count = (*request)->size;
  status->MPI_SOURCE = (*request)->peer;
  status->MPI_TAG = (*request)->tag;
  if ((*request)->active == 2) {
    (*request)->active = 0;     // modified by Qing
  } else {
    _USF_request_remove ((*request)->node);
    *request = MPI_REQUEST_NULL;
  }

  pthread_mutex_unlock (&_USF_mut_req);
  return 0;
}
